# -*- coding: utf-8 -*-

import os
import optparse
import sys

import mamonsu.lib.platform as platform
from mamonsu.plugins.pgsql.driver.checks import is_conn_to_db
from mamonsu import __version__ as mamonsu_version
from mamonsu.lib.default_config import DefaultConfig
from mamonsu.plugins.pgsql.pool import Pooler
from mamonsu.tools.bootstrap.sql import CreateSchemaSQL, GrantsOnSchemaSQL, QuerySplit


class Args(DefaultConfig):

    def __init__(self):
        parser = optparse.OptionParser(
            usage='%prog bootstrap -M <MAMONSU USERNAME> <DBNAME>',
            version='%prog bootstrap {0}'.format(mamonsu_version),
            description='Bootstrap DDL for monitoring')
        group = optparse.OptionGroup(
            parser,
            'Postgres connection options')
        group.add_option(
            '-d', '--dbname',
            dest='dbname',
            default=self.default_db(),
            help='database name to connect to (default: %default)')
        group.add_option(
            '--host',
            dest='hostname',
            default=self.default_host(),
            help='database server host or socket path (default: %default)')
        group.add_option(
            '--port',
            dest='port',
            default=self.default_port(),
            help='database server port (default: %default)')
        group.add_option(
            '-U', '--username',
            dest='username',
            default=self.default_user(),
            help='database superuser name (default: %default)')
        group.add_option(
            '-W', '--password',
            dest='password',
            default=self.default_user())
        bootstrap_group = optparse.OptionGroup(
            parser,
            'Bootstrap options')
        bootstrap_group.add_option(
            '-v', '--verbose',
            action="store_true",
            dest="verbose",
            default=False,
            help='Show bootstrap DDL')
        bootstrap_group.add_option(
            '-M', '--mamonsu-username',
            dest='mamonsu_username',
            help='database non-priviledged user for mamonsu')
        parser.add_option_group(group)
        parser.add_option_group(bootstrap_group)

        self.args, commands = parser.parse_args()

        if len(commands) > 0:
            if len(commands) == 1:
                self.args.dbname = commands[0]
            else:
                parser.print_help()
                sys.exit(1)
        if self.args.mamonsu_username is None:
            sys.stderr.write("ERROR: Database non-priviledged username for mamonsu is not specified\n")
            parser.print_help()
            sys.exit(1)

        # apply env
        os.environ['PGUSER'] = self.args.username
        os.environ['PGPASSWORD'] = self.args.password
        os.environ['PGHOST'] = self.args.hostname
        os.environ['PGDATABASE'] = self.args.dbname
        os.environ['PGAPPNAME'] = 'mamonsu deploy'

    def try_configure_connect_to_pg(self):
        if not self._configure_auto_host():
            if self._try_run_as_postgres():
                if not self._configure_auto_host():
                    sys.stderr.write(
                        "Can't connect as user postgres,"
                        " may be database settings wrong?\n")
                    return False
            else:
                sys.stderr.write(
                    "Can't connect with host=auto,"
                    " may be database settings wrong?\n")
                return False
        else:
            return True

    def _try_run_as_postgres(self):
        if platform.UNIX and os.getegid() == 0:
            try:
                import pwd
                uid = pwd.getpwnam('postgres').pw_uid
                os.seteuid(uid)
                return True
            except Exception as e:
                sys.stderr.write("Failed run as postgres: {0}\n".format(e))
                pass
        return False

    def _configure_auto_host(self):

        def test_db(self, host_pre):
            if is_conn_to_db(
                host=host_pre,
                db=self.args.dbname,
                port=self.args.port,
                user=self.args.username,
                    paswd=self.args.password):
                self.args.hostname = host_pre
                os.environ['PGHOST'] = self.args.hostname
                return True
            return False

        host = self.args.hostname
        port = self.args.port
        if host == 'auto' and platform.UNIX:
            if test_db(self, '/tmp/.s.PGSQL.{0}'.format(port)):
                return True
            if test_db(self, '/var/run/postgresql/.s.PGSQL.{0}'.format(port)):
                return True
            if test_db(self, '127.0.0.1'):
                return True
            # auto failed
            return False
        # not auto
        return True


def run_deploy():

    args = Args()
    if not args.try_configure_connect_to_pg():
        sys.exit(1)

    if not Pooler.is_superuser():
        sys.stderr.write(
            "ERROR: Bootstrap must be run by PostgreSQL superuser\n")
        sys.exit(1)

    try:
        for sql in CreateSchemaSQL.format(
                mamonsu_version,
                mamonsu_version.replace('.', '_'),
                '[0-9A-F]{24}',
                'wal' if Pooler.server_version_greater('10.0') else 'xlog',
                'wal_lsn' if Pooler.server_version_greater('10.0') else 'xlog_location'
                ).split(QuerySplit):
            if args.args.verbose:
                sys.stdout.write("\nExecuting query:\n{0}\n".format(sql))
            Pooler.query(sql)
    except Exception as e:
        sys.stderr.write("Query:\n{0}\nerror: {1}\n".format(sql, e))
        sys.exit(2)

    try:
        for sql in GrantsOnSchemaSQL.format(
                mamonsu_version.replace('.', '_'),
                args.args.mamonsu_username,
                'wal' if Pooler.server_version_greater('10.0') else 'xlog'
                ).split(QuerySplit):
            if args.args.verbose:
                sys.stdout.write("\nExecuting query:\n{0}\n".format(sql))
            Pooler.query(sql)
    except Exception as e:
        sys.stderr.write("Query:\n{0}\nerror: {1}\n".format(sql, e))
        sys.exit(2)

    sys.stdout.write("Bootstrap successfully completed\n")



import sys

LINUX = (sys.platform == 'linux' or sys.platform == 'linux2')
WINDOWS = (sys.platform == 'win32' or sys.platform == 'win64')
FREEBSD = ('freebsd' in sys.platform)
UNIX = LINUX or FREEBSD

PY2 = (sys.version_info[0] == 2)
PY3 = (sys.version_info[0] == 3)

if PY2:
    INTEGER_TYPES = (int, long)
if PY3:
    INTEGER_TYPES = int,



# -*- coding: utf-8 -*-

# https://www.zabbix.com/documentation/2.0/ru/manual/appendix/items/activepassive

import time
import struct
import socket
import json
import logging
import encodings.idna

import mamonsu.lib.platform as platform
from mamonsu.lib.plugin import Plugin
from mamonsu.lib.queue import Queue


class ZbxSender(Plugin):

    Interval = 10
    _sender = True

    def __init__(self, config):
        super(ZbxSender, self).__init__(config)
        self.host = config.fetch('zabbix', 'address')
        if self.host is None:
            self._enabled = False
        elif not config.fetch('zabbix', 'enabled', bool):
            self._enabled = False
        self.port = config.fetch('zabbix', 'port', int)
        self.max_queue_size = config.fetch('sender', 'queue', int)
        self.fqdn = config.fetch('zabbix', 'client')
        self.queue = Queue()
        self.log = logging.getLogger(
            'ZBX-{0}:{1}'.format(self.host, self.port))

    def send(self, key, value, host=None, clock=None):
        if host is None:
            host = self.fqdn
        if clock is None:
            clock = int(time.time())
        metric = {
            'host': host, 'key': key,
            'value': str(value), 'clock': clock}
        self._send(metric)

    def _send(self, metric):
        if self.queue.size() > self.max_queue_size:
            self.log.error('Queue size over limit, replace last metric')
            self.queue.replace(metric)
        else:
            self.queue.add(metric)

    def run(self, zbx):
        self._flush()

    def _flush(self):
        metrics = self.queue.flush()
        if len(metrics) == 0:
            return
        data = json.dumps({
            'request': 'sender data',
            'data': metrics,
            'clock': int(time.time())
        })
        self._send_data(data)

    def _send_data(self, data):
        data_len = struct.pack('<Q', len(data))
        if platform.PY3:
            packet = b'ZBXD\x01' + data_len + str.encode(data)
        else:
            packet = 'ZBXD\x01' + data_len + data
        try:
            sock = socket.socket()
            sock.connect((self.host, self.port))
            self.log.debug('request: {0}'.format(data))
            sock.sendall(packet)
            resp_header = self._receive(sock, 13)
            resp_body_len = struct.unpack('<Q', resp_header[5:])[0]
            resp_body = self._receive(sock, resp_body_len)
            self.log.debug('response: {0}'.format(resp_body))
            if 'failed: 0' not in str(resp_body):
                self.log.error(
                    'On request:\n{0}\nget response'
                    ' with failed items:\n{1}'.format(
                        data,
                        resp_body))
        finally:
            sock.close()

    def _receive(self, sock, count):
        buf = str.encode('')
        while len(buf) < count:
            chunk = sock.recv(count - len(buf))
            if not chunk:
                break
            buf += chunk
        return buf



# -*- coding: utf-8 -*-
import os
import threading
import logging

from mamonsu.plugins.pgsql.driver.pg8000 import connect
from mamonsu.plugins.pgsql.driver.pg8000.core import ProgrammingError


class ConnectionInfo(object):

    def __init__(self, connection_hash={}):
        self.host = connection_hash.get('host') or os.environ.get('PGHOST')
        self.port = connection_hash.get('port') or int(os.environ.get('PGPORT') or 5432)
        self.user = connection_hash.get('user') or os.environ.get('PGUSER')
        self.passwd = connection_hash.get('passwd') or os.environ.get('PGPASSWORD')
        self.db = connection_hash.get('db') or os.environ.get('PGDATABASE')
        self.timeout = connection_hash.get('timeout') or int(
            os.environ.get('PGTIMEOUT') or 1)
        self.appname = connection_hash.get('appname') or os.environ.get('PGAPPNAME')
        self.log = logging.getLogger('PGSQL-({0})'.format(self.to_string()))

    def to_string(self):
        return 'host={0} db={1} user={2} port={3}'.format(
            self.host, self.db, self.user, self.port)

    def get_hash(self):
        return {
            'host': self.host,
            'port': self.port,
            'user': self.user,
            'passwd': self.passwd,
            'db': self.db,
            'timeout': self.timeout,
            'appname': self.appname
        }


class Connection(ConnectionInfo):

    def __init__(self, info={}):
        super(Connection, self).__init__(info)
        self.lock = threading.Lock()
        self.conn = None
        self.connected = False

    def query(self, query):
        self.lock.acquire()
        try:
            self.log.debug('Run: "{0}"'.format(query))
            self._check_connect()
            self.connected = False
            cursor = self.conn.cursor()
            cursor.execute(query)
            result = cursor.fetchall()
            cursor.close()
            self.connected = True
        except ProgrammingError as e:
            error_text = '{0}'.format(e)
            if error_text == 'no result set':
                return None
            else:
                raise ProgrammingError(error_text)
        finally:
            self.lock.release()
        return result

    def _close(self):
        if self.conn is not None:
            self.log.debug('closing old connection')
            try:
                self.conn.close()
            except Exception as e:
                self.log.error('close error: {0}'.format(e))

    def _connect(self):
        self.log.debug('connecting')
        host, unix_sock = self.host, None
        if host.startswith('/'):
            unix_sock, host = host, None
        self.conn = connect(
            user=self.user,
            password=self.passwd,
            unix_sock=unix_sock,
            host=host,
            port=self.port,
            database=self.db,
            application_name=self.appname
        )
        self.log.debug('connected')
        self.conn.autocommit = True
        cur = self.conn.cursor()
        cur.execute('set statement_timeout to {0}'.format(self.timeout * 1000))
        cur.close()
        self.log.debug('ready')

    def _check_connect(self):
        if not self.connected:
            self.log.debug('reconnecting')
            self._close()
            self._connect()



# -*- coding: utf-8 -*-

from mamonsu.plugins.pgsql.plugin import PgsqlPlugin as Plugin
from .pool import Pooler


class Oldest(Plugin):

    OldestXidSql = """
select
    greatest(max(age(backend_xmin)), max(age(backend_xid)))
from pg_catalog.pg_stat_activity;
"""

    OldestXidSql_bootstrap = """
select public.mamonsu_get_oldest_xid();
"""

    OldestQuerySql = """
select
    extract(epoch from max(now() - xact_start))
from pg_catalog.pg_stat_activity;
"""

    OldestQuerySql_bootstrap = """
select public.mamonsu_get_oldest_query();
"""

    DEFAULT_CONFIG = {
        'max_xid_age': str(5000 * 60 * 60),
        'max_query_time': str(5 * 60 * 60)
    }

    def run(self, zbx):
        if Pooler.is_bootstraped() and Pooler.bootstrap_version_greater('2.3.2'):
            xid = Pooler.query(self.OldestXidSql_bootstrap)[0][0]
            query = Pooler.query(self.OldestQuerySql_bootstrap)[0][0]
        else:
            xid = Pooler.query(self.OldestXidSql)[0][0]
            query = Pooler.query(self.OldestQuerySql)[0][0]

        zbx.send('pgsql.oldest[xid_age]', xid)
        zbx.send('pgsql.oldest[query_time]', query)

    def graphs(self, template):
        result = template.graph({
            'name': 'PostgreSQL oldest query running time',
            'items': [{
                'key': 'pgsql.oldest[query_time]',
                'color': '00CC00'
            }]
        })
        result += template.graph({
            'name': 'PostgreSQL age of oldest xid',
            'items': [{
                'key': 'pgsql.oldest[xid_age]',
                'color': '00CC00'
            }]
        })
        return result

    def items(self, template):
        return template.item({
            'key': 'pgsql.oldest[xid_age]',
            'name': 'PostgreSQL: age of oldest xid',
            'value_type': Plugin.VALUE_TYPE.numeric_unsigned
        }) + template.item({
            'key': 'pgsql.oldest[query_time]',
            'name': 'PostgreSQL: oldest query running time in sec',
            'units': Plugin.UNITS.s
        })

    def triggers(self, template):
        return template.trigger({
            'name': 'PostgreSQL oldest xid is too big on {HOSTNAME}',
            'expression': '{#TEMPLATE:pgsql.oldest[xid_age]'
            '.last()}&gt;' + self.plugin_config('max_xid_age')
        }) + template.trigger({
            'name': 'PostgreSQL query running is too old on {HOSTNAME}',
            'expression': '{#TEMPLATE:pgsql.oldest[query_time]'
            '.last()}&gt;' + self.plugin_config('max_query_time')
        })



# -*- coding: utf-8 -*-

from mamonsu.plugins.pgsql.plugin import PgsqlPlugin as Plugin
from .pool import Pooler


class Checkpoint(Plugin):

    Interval = 60 * 5

    DEFAULT_CONFIG = {'max_checkpoint_by_wal_in_hour': str(12)}

    Items = [
        # key, zbx_key, description,
        #    ('graph name', color, side), units, delta, factor

        ('checkpoints_timed', 'checkpoint[count_timed]',
            'checkpoints: by timeout (in hour)',
            ('PostgreSQL checkpoint', '00CC00', 0),
            Plugin.UNITS.none, Plugin.DELTA.speed_per_second, 60 * 60),

        ('checkpoints_req', 'checkpoint[count_wal]',
            'checkpoints: by wal (in hour)',
            ('PostgreSQL checkpoint', 'CC0000', 0),
            Plugin.UNITS.none, Plugin.DELTA.speed_per_second, 60 * 60),

        ('checkpoint_write_time', 'checkpoint[write_time]',
            'checkpoint: write time',
            ('PostgreSQL checkpoints', '0000CC', 1),
            Plugin.UNITS.ms, Plugin.DELTA.speed_per_second, 1),

        ('checkpoint_sync_time', 'checkpoint[checkpoint_sync_time]',
            'checkpoint: sync time',
            ('PostgreSQL checkpoints', '000000', 1),
            Plugin.UNITS.ms, Plugin.DELTA.speed_per_second, 1)
    ]

    def run(self, zbx):
        params = [x[0] for x in self.Items]
        result = Pooler.query(
            'select {0} from pg_catalog.pg_stat_bgwriter'.format(
                ', '.join(params)))
        for idx, val in enumerate(result[0]):
            key, val = 'pgsql.{0}'.format(
                self.Items[idx][1]), int(val)
            zbx.send(key, val * self.Items[idx][6], self.Items[idx][5])
        del params, result

    def items(self, template):
        result = ''
        for item in self.Items:
            result += template.item({
                'key': 'pgsql.{0}'.format(item[1]),
                'name': 'PostgreSQL {0}'.format(item[2]),
                'value_type': Plugin.VALUE_TYPE.numeric_float,
                'units': item[4],
                'delay': self.Interval
            })
        return result

    def graphs(self, template):
        name, items = 'PostgreSQL checkpoints', []
        for item in self.Items:
            items.append({
                'key': 'pgsql.{0}'.format(item[1]),
                'color': item[3][1],
                'yaxisside': item[3][2],
                'delay': self.Interval
            })
        return template.graph({'name': name, 'items': items})

    def triggers(self, template):
        return template.trigger({
            'name': 'PostgreSQL required checkpoints occurs to '
            'frequently on {HOSTNAME}',
            'expression': '{#TEMPLATE:pgsql.checkpoint[count_wal]'
            '.last()}&gt;' + self.plugin_config(
                'max_checkpoint_by_wal_in_hour')
        })



# -*- coding: utf-8 -*-

from mamonsu.plugins.pgsql.plugin import PgsqlPlugin as Plugin
from .pool import Pooler


class Connections(Plugin):

    # (state, key, name, graph)
    Items = [
        ('active', 'active', 'number of active connections', '00BB00'),
        ('idle', 'idle', 'number of idle connections', '0000BB'),
        ('idle in transaction', 'idle_in_transaction',
            'number of idle in transaction connections', 'CC00CC')
    ]

    def run(self, zbx):

        if Pooler.is_bootstraped() and Pooler.bootstrap_version_greater('2.3.4'):
            result = Pooler.query(
                'select state, count(*) '
                'from mamonsu_get_connections_states() group by state')
        else:
            result = Pooler.query(
                'select state, count(*) '
                'from pg_catalog.pg_stat_activity group by state')
        for item in self.Items:
            state, key, val = item[0], item[1], 0
            for row in result:
                if row[0] != state:
                    continue
                else:
                    val = row[1]
                    break
            zbx.send('pgsql.connections[{0}]'.format(key), float(val))

        if Pooler.is_bootstraped() and Pooler.bootstrap_version_greater('2.3.4'):
            result = Pooler.query(
                'select count(*) '
                'from mamonsu_get_connections_states()')
        else:
            result = Pooler.query(
                'select count(*) '
                'from pg_catalog.pg_stat_activity')

        zbx.send('pgsql.connections[total]', int(result[0][0]))

        if Pooler.server_version_less('9.5.0'):
            if Pooler.is_bootstraped() and Pooler.bootstrap_version_greater('2.3.4'):
                result = Pooler.query(
                    'select count(*) '
                    'from mamonsu.mamonsu_get_connections_states() '
                    'where waiting')
            else:
                result = Pooler.query(
                    'select count(*) '
                    'from pg_catalog.pg_stat_activity where waiting')
            zbx.send('pgsql.connections[waiting]', int(result[0][0]))

    def items(self, template):
        result = template.item({
            'name': 'PostgreSQL: number of total connections',
            'key': 'pgsql.connections[total]'
        })
        result += template.item({
            'name': 'PostgreSQL: number of waiting connections',
            'key': 'pgsql.connections[waiting]'
        })
        for item in self.Items:
            result += template.item({
                'name': 'PostgreSQL: {0}'.format(item[2]),
                'key': 'pgsql.connections[{0}]'.format(item[1])
            })
        return result

    def graphs(self, template):
        items = []
        for item in self.Items:
            items.append({
                'key': 'pgsql.connections[{0}]'.format(item[1]),
                'color': item[3]
            })
        items.append({
            'key': 'pgsql.connections[total]',
            'color': 'EEEEEE'
        })
        items.append({
            'key': 'pgsql.connections[waiting]',
            'color': 'BB0000'
        })
        graph = {'name': 'PostgreSQL connections', 'items': items}
        return template.graph(graph)



# -*- coding: utf-8 -*-
import sys
import optparse

from mamonsu import __version__
from mamonsu.lib.const import API
from mamonsu.lib.config import Config
import mamonsu.lib.platform as platform

if platform.PY3:
    import urllib.request as urllib2
else:
    import urllib2


def run_agent():

    def print_help():
        print(usage_msg)
        sys.exit(7)

    usage_msg = """
{prog} agent [OPTIONS] COMMANDS

Examples:
    {prog} agent version
    {prog} agent metric-list
    {prog} agent metric-get <metric key>
""".format(prog=sys.argv[0])

    parser = optparse.OptionParser(
        usage=usage_msg,
        version='%prog agent {0}'.format(__version__))
    parser.add_option(
        '-c', '--config', dest='config',
        help=optparse.SUPPRESS_HELP)
    args, commands = parser.parse_args()

    cfg = Config(args.config)
    host = cfg.fetch('agent', 'host')
    port = cfg.fetch('agent', 'port', int)
    url, key = None, None

    if len(commands) == 0:
        return print_help()
    if commands[0] == 'version':
        if len(commands) >= 2:
            return print_help()
        url = 'http://{0}:{1}/version'.format(host, port)
    elif commands[0] == 'metric-list':
        if len(commands) >= 2:
            return print_help()
        url = 'http://{0}:{1}/list'.format(host, port)
    elif commands[0] == 'metric-get':
        if not len(commands) == 2:
            return print_help()
        key = commands[1]
        url = 'http://{0}:{1}/get?key={2}'.format(host, port, key)
    else:
        return print_help()

    request = urllib2.Request(url)
    try:
        response = urllib2.urlopen(request)
    except Exception as e:
        sys.stderr.write('Open url: {0}, error: {1}\n'.format(
            url, e))
        sys.exit(9)

    if not response.code == 200:
        sys.stderr.write('Bad response from url {0}, code: {1}\n'.format(
            url,
            response.code))
        sys.exit(8)
    else:
        try:
            body = response.read()
        except Exception as e:
            sys.stderr.write('Read url: {0}, error: {1}\n'.format(
                url, e))
            sys.exit(9)
        if body == API.UNKNOWN_VERSION:
            sys.stderr.write('Unknown API version\n')
            sys.exit(9)
        elif body == API.METRIC_NOT_FOUND and key is not None:
            sys.stderr.write('Metric \'{0}\' not found\n'.format(key))
            sys.exit(9)
        elif body == API.METRIC_NOT_FOUND and key is None:
            sys.stderr.write('Unknown API version\n')
            sys.exit(9)
        else:
            print(body)



# -*- coding: utf-8 -*-

import re
import sys
import mamonsu.lib.platform as platform


class color(object):

    mapping = {
        'BOLD': '\033[0;0m\033[1;1m',
        'RED': '\033[1;31m',
        'GRAY': '\033[1;30m',
        'PURPLE': '\033[1;35m',
        'BLUE': '\033[1;34m',
        'END': '\033[1;m'
    }

    def __init__(self):
        self.color = sys.stdout.isatty()

    def disable(self):
        self.color = False

    def __getattr__(self, name):
        if self.color:
            return self.mapping[name]
        else:
            return ''


TermColor = color()


# int (bytes) => str (human readable)
def humansize_bytes(nbytes):
    fmt = '{0:>6} {1}'
    if not isinstance(nbytes, platform.INTEGER_TYPES):
        return 'ERROR'
    if nbytes == 0:
        return fmt.format(0, 'B')
    i, suffixes, = 0, ['B', 'KB', 'MB', 'GB', 'TB', 'PB']
    while nbytes >= 1024 and i < len(suffixes) - 1:
        nbytes /= 1024.
        i += 1
    f = ('%.2f' % nbytes).rstrip('0').rstrip('.')
    return fmt.format(f, suffixes[i])


# str (some formates) => str (human readable)
def humansize(value):
    m = re.search('(\d+) (\S+)', value)
    if m is None:
        return value
    val, suff = m.group(1), m.group(2)
    val, suff = int(val), suff.upper()
    if suff == 'S':
        return value
    if suff == 'MS':
        return value
    if suff == 'B':
        return humansize_bytes(val)
    if suff == 'KB':
        return humansize_bytes(val * 1024)
    if suff == '4KB':
        return humansize_bytes(val * 1024 * 4)
    if suff == '8KB':
        return humansize_bytes(val * 1024 * 8)
    if suff == '16KB':
        return humansize_bytes(val * 1024 * 16)
    if suff == 'MB':
        return humansize_bytes(val * 1024 * 1024)
    if suff == 'GB':
        return humansize_bytes(val * 1024 * 1024 * 1024)
    if suff == 'TB':
        return humansize_bytes(val * 1024 * 1024 * 1024 * 1024)
    return value


def header_h1(info):
    return "\n{0}{1}{2}{3}\n".format(
        TermColor.BOLD, TermColor.RED, info.upper(), TermColor.END)


def key_val_h1(key, val, spaces=12):
    fmt = "  {0}{1}{2:" + str(spaces) + "}{3}: {4}\n"
    return fmt.format(
        TermColor.BOLD, TermColor.PURPLE, key, TermColor.END, val)


def header_h2(info):
    return "  {0}{1}{2}{3}\n".format(
        TermColor.BOLD, TermColor.PURPLE, info, TermColor.END)


def key_val_h2(key, val, delim=': '):
    return "    {0}{1}{2:4}{3}{4}{5}\n".format(
        TermColor.BOLD, TermColor.BLUE, key, TermColor.END, delim, val)


def topline_h1(arr=[], delim="  \t"):
    result = "{0}{1}".format(TermColor.BOLD, TermColor.BLUE)
    for x in arr:
        result = "{0}{1}{2}".format(result, delim, x)
    return "{0}{1}\n".format(result, TermColor.END)


def format_raw_h1(raw=""):
    result = []
    for i, line in enumerate(raw.split("\n")):
        if i == 0:
            result.append("  {0}{1}{2}{3}".format(
                TermColor.BOLD, TermColor.BLUE, line, TermColor.END))
        else:
            result.append("  {0}".format(line))
    return "\n".join(result) + "\n"



# -*- coding: utf-8 -*-
from mamonsu.lib.plugin import Plugin

from mamonsu.lib.const import API
import mamonsu.lib.platform as platform

if platform.PY3:
    from urllib.parse import urlparse as parse
    from urllib.parse import parse_qs
    from http.server import BaseHTTPRequestHandler, HTTPServer
else:
    from urlparse import urlparse as parse
    from urlparse import parse_qs
    from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer


class AgentApi(Plugin):

    def __init__(self, config):
        super(AgentApi, self).__init__(config)
        self._enabled = config.fetch('agent', 'enabled', bool)
        self.host = config.fetch('agent', 'host')
        self.port = config.fetch('agent', 'port', int)

    def run(self, _=None):
        self.log.info('Starting at http://{0}:{1}'.format(
            self.host, self.port))
        self.server = AgentApiHttpServer(self)


class AgentApiHttpServer:

    def __init__(self, config):
        def handler(*args):
            AgentApiHandler(config, *args)
        server = HTTPServer((config.host, config.port), handler)
        server.serve_forever()


class AgentApiHandler(BaseHTTPRequestHandler):

    def __init__(self, config, *args):
        self.sender = config.sender
        BaseHTTPRequestHandler.__init__(self, *args)

    def do_HEAD(self):
        self.send_response(200)
        self.send_header('Content-type', 'text/html')
        self.end_headers()

    def do_GET(self):
        req = parse(self.path)
        if req.path in '/version':
            self.wfile.write(API.VERSION)
        # get metric
        elif req.path in API.METRIC_GET_URLS:
            query = parse_qs(req.query)
            key, host = None, None
            if 'key' in query:
                key = query['key'][0]
            if 'host' in query:
                host = query['host'][0]
            resp = self.sender.get_metric(key, host)
            if resp[0] is not None:
                result = '{0}\t{1}\t{2}'.format(
                    key, resp[0], resp[1])
                if platform.PY3:
                    result = bytearray(result, 'utf-8')
                self.wfile.write(result)
            else:
                self.wfile.write(API.METRIC_NOT_FOUND)
        # get list of metric
        elif req.path in API.METRIC_LIST_URLS:
            query, host, result = parse_qs(req.query), None, ''
            if 'host' in query:
                host = query['host'][0]
            for val in self.sender.list_metrics(host):
                result += '{0}\t{1}\t{2}\n'.format(
                    val[0], val[1][0], val[1][1])
            if platform.PY3:
                result = bytearray(result, 'utf-8')
            self.wfile.write(result)
        else:
            # unknown path
            self.wfile.write(API.UNKNOWN_VERSION)



# -*- coding: utf-8 -*-

import os
import logging
import optparse

from mamonsu import __version__
from mamonsu.lib.default_config import DefaultConfig
from mamonsu.tools.zabbix_cli.operations import Operations


class Args(DefaultConfig):

    def __init__(self):

        parser = optparse.OptionParser(
            usage='%prog zabbix [--url] [--user] '
            '[--password] [--log-level] commands',
            version='%prog zabbix cli tools {0}'.format(__version__),
            description='Zabbix CLI tools')
        group = optparse.OptionGroup(
            parser,
            'Zabbix API connection settings')
        group.add_option(
            '--url',
            dest='zabbix_url',
            default=os.environ.get('ZABBIX_URL') or 'http://localhost/zabbix',
            help='URL of Zabbix Web interface (default: %default)')
        group.add_option(
            '--user',
            dest='zabbix_user',
            default=os.environ.get('ZABBIX_USER') or 'Admin',
            help='User (default: %default)')
        group.add_option(
            '--password',
            dest='zabbix_password',
            default=os.environ.get('ZABBIX_PASSWD') or 'zabbix',
            help='Password (default: %default)')
        group.add_option(
            '-l', '--log-level',
            dest='log_level',
            default='INFO', help='Log level (default: %default)')
        parser.add_option_group(group)

        self.args, commands = parser.parse_args()
        self.args.commands = commands

        # apply logging
        logging.basicConfig(
            level=self.get_logger_level(self.args.log_level))

    def __getattr__(self, name):
        try:
            return self.args.__dict__[name]
        except KeyError:
            return None


def run_zabbix():
    args = Args()
    Operations(args)



from mamonsu.plugins.system.plugin import SystemPlugin as Plugin
from operator import add
from math import pow
import re


class Mem_Frag(Plugin):

    def parse_line(self, line):
        line = line.strip()
        parsed_line = re.match("Node\s+(?P<numa_node>\d+).*zone\s+(?P<zone>\w+)\s+(?P<nr_free>.*)", line).groupdict()
        return parsed_line

    def get_per_node_hash(self):
        max_node = 0
        buddyhash = {}
        buddyinfo = open("/proc/buddyinfo").readlines()
        for line in map(self.parse_line, buddyinfo):
                numa_node =  int(line["numa_node"])
                free_fragments = map(int, line["nr_free"].split())
                if numa_node not in buddyhash:
                    buddyhash[numa_node] = free_fragments
                else:
                    buddyhash[numa_node] = map(add, buddyhash[numa_node], free_fragments)
        return buddyhash

    def calc_fraglevel(self):
        combined_hash = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        fraglevel = []
        buddyhash = self.get_per_node_hash()
        for node in buddyhash:
            combined_hash = map(add, buddyhash[node], combined_hash)
        total_free_pages = 0
        for i in range(len(combined_hash)):
            total_free_pages = total_free_pages + int(combined_hash[i]*pow(2, i))

        # fraglevel_for_specific_order = (total_free_pages -sum_free_pages_of_equal_and_senior_orders*100/ total_free_pages)
        # fragmentation is a ratio of how many of available free pages cannot be used for allocations of your order

        for i in range(len(combined_hash)):
            pages_of_senior_and_equal_orders = 0
            for n in range(i, len(combined_hash)):
                pages_of_senior_and_equal_orders = pages_of_senior_and_equal_orders + int(combined_hash[n]*pow(2, n))
            fraglevel_for_specific_order = (float(total_free_pages - pages_of_senior_and_equal_orders)*100)/float(total_free_pages)
            fraglevel.append(fraglevel_for_specific_order)
        total_fraglevel = float(sum(fraglevel))/float(len(fraglevel))
        return total_fraglevel

    def run(self, zbx):
        mem_frag = self.calc_fraglevel()
        zbx.send('system.mem_frag[1]', float(mem_frag))

    def items(self, template):
        return template.item({
            'name': 'System wide memory fragmentation',
            'key': 'system.mem_frag[1]'
        })

    def graphs(self, template):
        items = [{'key': 'system.mem_frag[1]'}]
        graph = {'name': 'System memory fragmentation', 'items': items}
        return template.graph(graph)



from mamonsu import __version__ as mamonsu_version

QuerySplit = """

"""

CreateSchemaSQL = """
CREATE TABLE IF NOT EXISTS public.mamonsu_config (
  version text,
  inserted_at timestamp DEFAULT NOW()
);

INSERT INTO public.mamonsu_config(version) VALUES('{0}');

DROP TABLE IF EXISTS public.mamonsu_timestamp_master_{1};

CREATE TABLE public.mamonsu_timestamp_master_{1}(
    id int primary key,
    ts double precision,
    lsn pg_lsn
);

INSERT INTO public.mamonsu_timestamp_master_{1} (id) values (1);

CREATE OR REPLACE FUNCTION public.mamonsu_timestamp_master_update()
RETURNS void AS $$
  UPDATE public.mamonsu_timestamp_master_{1} SET
    ts = extract(epoch from now() at time zone 'utc')::double precision,
    lsn = pg_catalog.pg_current_{4}()
  WHERE
    id = 1;
$$ LANGUAGE SQL SECURITY DEFINER;

CREATE OR REPLACE FUNCTION public.mamonsu_timestamp_get()
RETURNS double precision AS $$
  SELECT
    (extract(epoch from now() at time zone 'utc') - ts)::double precision
  FROM public.mamonsu_timestamp_master_{1}
  WHERE id = 1 LIMIT 1;
$$ LANGUAGE SQL SECURITY DEFINER;

CREATE OR REPLACE FUNCTION public.mamonsu_count_autovacuum()
RETURNS BIGINT AS $$
    SELECT
        count(*)::BIGINT
    FROM pg_catalog.pg_stat_activity
    WHERE
        query like '%%autovacuum%%'
        and state <> 'idle'
        and pid <> pg_catalog.pg_backend_pid()
$$ LANGUAGE SQL SECURITY DEFINER;

CREATE OR REPLACE FUNCTION public.mamonsu_get_connections_states()
RETURNS TABLE(state text) AS $$
    SELECT
        state
    FROM pg_catalog.pg_stat_activity
$$ LANGUAGE SQL SECURITY DEFINER;

CREATE or REPLACE FUNCTION public.mamonsu_get_oldest_xid()
RETURNS BIGINT AS $$
    SELECT
        greatest(max(age(backend_xmin)),
        max(age(backend_xid)))::BIGINT
    FROM pg_catalog.pg_stat_activity
$$ LANGUAGE SQL SECURITY DEFINER;

CREATE or REPLACE FUNCTION public.mamonsu_get_oldest_query()
RETURNS DOUBLE PRECISION AS $$
    SELECT
        extract(epoch from max(now() - xact_start))
    FROM pg_catalog.pg_stat_activity
$$ LANGUAGE SQL SECURITY DEFINER;

CREATE OR REPLACE FUNCTION public.mamonsu_count_{3}_files()
RETURNS BIGINT AS $$
WITH list(filename) as (SELECT * FROM pg_catalog.pg_ls_dir('pg_{3}'))
SELECT
    COUNT(*)::BIGINT
FROM
    list
WHERE filename similar to '{2}'
$$ LANGUAGE SQL SECURITY DEFINER;

CREATE EXTENSION IF NOT EXISTS pg_buffercache;

CREATE OR REPLACE FUNCTION public.mamonsu_buffer_cache()
RETURNS TABLE(SIZE BIGINT, TWICE_USED BIGINT, DIRTY BIGINT) AS $$
SELECT
   SUM(1) * 8 * 1024,
   SUM(CASE WHEN usagecount > 1 THEN 1 ELSE 0 END) * 8 * 1024,
   SUM(CASE isdirty WHEN true THEN 1 ELSE 0 END) * 8 * 1024
FROM public.pg_buffercache
$$ LANGUAGE SQL SECURITY DEFINER;
"""

GrantsOnSchemaSQL = """
ALTER TABLE public.mamonsu_config OWNER TO {1};

ALTER TABLE public.mamonsu_timestamp_master_{0} OWNER TO {1};

GRANT EXECUTE ON FUNCTION public.mamonsu_timestamp_master_update() TO {1};

GRANT EXECUTE ON FUNCTION public.mamonsu_timestamp_get() TO {1};

GRANT EXECUTE ON FUNCTION public.mamonsu_count_autovacuum() TO {1};

GRANT EXECUTE ON FUNCTION public.mamonsu_get_oldest_xid() TO {1};

GRANT EXECUTE ON FUNCTION public.mamonsu_get_oldest_query() TO {1};

GRANT EXECUTE ON FUNCTION public.mamonsu_count_{2}_files() TO {1};

GRANT EXECUTE ON FUNCTION public.mamonsu_buffer_cache() TO {1};
"""



import re
from mamonsu.plugins.system.plugin import SystemPlugin as Plugin


class DiskStats(Plugin):

    # todo yaxis right 100%
    # bold line

    # Track only physical devices without logical partitions
    OnlyPhysicalDevices = True

    re_stat = re.compile('^(?:\s+\d+){2}\s+([\w\d]+) (.*)$')
    # rd_ios rd_merges rd_sectors rd_ticks
    # wr_ios wr_merges wr_sectors wr_ticks
    # ios_in_prog tot_ticks rq_ticks

    def run(self, zbx):
        with open('/proc/diskstats', 'r') as f:

            devices = []
            all_read_op, all_read_b, all_write_op, all_write_b, = 0, 0, 0, 0

            for line in f:
                if re.search('(ram|loop)', line):
                    continue
                m = self.re_stat.match(line)
                if m is None:
                    continue
                dev, val = m.group(1), m.group(2)
                if self.OnlyPhysicalDevices and re.search('\d+$', dev):
                    continue
                val = [int(x) for x in val.split()]
                read_op, read_sc, write_op, write_sc, ticks = val[0], val[2], val[4], val[6], val[9]
                read_b, write_b = read_sc * 512, write_sc * 512  # https://github.com/sysstat/sysstat/blob/v11.5.2/iostat.c#L940

                zbx.send('system.disk.read[{0}]'.format(
                    dev), read_op, self.DELTA_SPEED)
                zbx.send('system.disk.write[{0}]'.format(
                    dev), write_op, self.DELTA_SPEED)
                zbx.send('system.disk.read_b[{0}]'.format(
                    dev), read_b, self.DELTA_SPEED)
                zbx.send('system.disk.write_b[{0}]'.format(
                    dev), write_b, self.DELTA_SPEED)
                zbx.send('system.disk.utilization[{0}]'.format(
                    dev), ticks / 10, self.DELTA_SPEED)

                all_read_op += read_op
                all_write_op += write_op
                all_read_b += read_b
                all_write_b += write_b
                devices.append({'{#BLOCKDEVICE}': dev})

            zbx.send('system.disk.all_read[]', all_read_op, self.DELTA_SPEED)
            zbx.send('system.disk.all_write[]', all_write_op, self.DELTA_SPEED)
            zbx.send('system.disk.all_read_b[]', all_read_b, self.DELTA_SPEED)
            zbx.send('system.disk.all_write_b[]', all_write_b, self.DELTA_SPEED)
            zbx.send('system.disk.discovery[]', zbx.json({'data': devices}))

    def items(self, template):
        return template.item({
            'name': 'Block devices: read requests',
            'key': 'system.disk.all_read[]'
        }) + template.item({
            'name': 'Block devices: write requests',
            'key': 'system.disk.all_write[]'
        }) + template.item({
            'name': 'Block devices: read byte/s',
            'key': 'system.disk.all_read_b[]'
        }) + template.item({
            'name': 'Block devices: write byte/s',
            'key': 'system.disk.all_write_b[]'
        })

    def graphs(self, template):
        graph = {
            'name': 'Block devices: read/write operations',
            'items': [
                {'key': 'system.disk.all_read[]', 'color': 'CC0000'},
                {'key': 'system.disk.all_write[]', 'color': '0000CC'}]
        }
        result = template.graph(graph)
        graph = {
            'name': 'Block devices: read/write bytes',
            'items': [
                {'key': 'system.disk.all_read_b[]', 'color': 'CC0000'},
                {'key': 'system.disk.all_write_b[]', 'color': '0000CC'}]
        }
        return result + template.graph(graph)

    def discovery_rules(self, template):

        rule = {
            'name': 'Block device discovery',
            'key': 'system.disk.discovery[]',
            'filter': '{#BLOCKDEVICE}:.*'
        }

        items = [
            {
                'key': 'system.disk.utilization[{#BLOCKDEVICE}]',
                'name': 'Block device {#BLOCKDEVICE}: utilization',
                'units': Plugin.UNITS.percent},
            {
                'key': 'system.disk.read[{#BLOCKDEVICE}]',
                'name': 'Block device {#BLOCKDEVICE}: read operations'},
            {
                'key': 'system.disk.write[{#BLOCKDEVICE}]',
                'name': 'Block device {#BLOCKDEVICE}: write operations'},
            {
                'key': 'system.disk.read_b[{#BLOCKDEVICE}]',
                'name': 'Block device {#BLOCKDEVICE}: read byte/s',
                'units': Plugin.UNITS.bytes},
            {
                'key': 'system.disk.write_b[{#BLOCKDEVICE}]',
                'name': 'Block device {#BLOCKDEVICE}: write byte/s',
                'units': Plugin.UNITS.bytes}]

        graphs = [{
            'name': 'Block device overview: {#BLOCKDEVICE} operations',
            'items': [{
                    'color': 'CC0000',
                    'key': 'system.disk.read[{#BLOCKDEVICE}]'},
                {
                    'color': '0000CC',
                    'key': 'system.disk.write[{#BLOCKDEVICE}]'},
                {
                    'yaxisside': 1,
                    'color': '00CC00',
                    'key': 'system.disk.utilization[{#BLOCKDEVICE}]'}]},
                  {
            'name': 'Block device overview: {#BLOCKDEVICE} byte/s',
            'items': [{
                    'color': 'CC0000',
                    'key': 'system.disk.read_b[{#BLOCKDEVICE}]'},
                {
                    'color': '0000CC',
                    'key': 'system.disk.write_b[{#BLOCKDEVICE}]'},
                {
                    'yaxisside': 1,
                    'color': '00CC00',
                    'key': 'system.disk.utilization[{#BLOCKDEVICE}]'}]
        }]

        return template.discovery_rule(rule=rule, items=items, graphs=graphs)
